#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

"""The :mod:`database <bibtexvcs.database>` module contains classes for managing a BibTeX VCS
database, which consists of a config file, a bib file, a documents directory, and a journals file.
"""
from __future__ import division, print_function, unicode_literals
import configparser, io, os, subprocess
from collections import OrderedDict
from os.path import join, exists, relpath

from bibtexvcs.bibfile import BibFile, MacroReference
from bibtexvcs.vcs import VCSInterface

BTVCSCONF = 'bibtexvcs.conf'  # name of the configuration file


class DatabaseFormatError(Exception):
    """Raised on any problem in the database layout (unparsable config-file, non-existing
    bib file, etc.).
    """
    pass


class Database:
    """Represents a complete literature database.

    :param directory: Base path of the database.
    :type directory: str

    The database configuration is read from the file `bbibtexvcs.conf` and from the layout of the
    directory that file resides in.

    .. attribute:: directory

        Absolute path of the root directory of the literature database.

    .. attribute:: bibfileName

        Name of the main bibtex database file.

    .. attribute:: bibfile

        :class:`BibFile` object parsed from the bibtex file.

    .. attribute:: journalsName

        Name of the journals file.

    .. attribute:: journals

        :class:`JournalsFile` object read from the journals file.

    .. attribute:: name

        Name of the database.

    .. attribute:: documents

        Name of the documents directory (relative to :attr:`directory`).

    .. attribute:: vcsType

        Type of the used VCS system. One of: ``"git"``, ``"mercurial"``, or ``None``
    """

    def __init__(self, directory):
        self.directory = directory
        if exists(join(self.directory, '.git')):
            self.vcsType = 'git'
        elif exists(join(self.directory, '.hg')):
            self.vcsType = 'mercurial'
        else:
            self.vcsType = None
        self._vcs = None
        self.reload()

    def reload(self):
        parser = configparser.ConfigParser()
        with io.open(self.configPath, encoding='UTF-8') as f:
            confFile = f.read()
        # workaround because ConfigParser does not support sectionless entries
        try:
            parser.read_string("[root]\n" + confFile)
        except configparser.Error as e:
            raise DatabaseFormatError("Could not parse configuration file '{}':\n\n{}"
                                      .format(BTVCSCONF, e))
        config = parser['root']

        self.bibfileName = config.get('bibfile', 'references.bib')
        self.journalsName = config.get('journals', 'journals.txt')

        for path in self.bibfilePath, self.journalsPath:
            if not exists(path):
                open(path, 'a').close()
                self.vcs.add(relpath(path, self.directory))

        self.bibfile = BibFile(join(self.directory, self.bibfileName))
        self.journals = JournalsFile(join(self.directory, self.journalsName))

        self.name = config.get('name', "Untitled Bibtex Database")
        self.documents = config.get('documents', 'Documents')

        self.makeJournalBibfiles()

        self.publicLink = config.get('publicLink', None)
        if not exists(self.documentsPath):
            os.mkdir(self.documentsPath)

    @property
    def journalsPath(self):
        """The absolute path of the `journals` file."""
        return join(self.directory, self.journalsName)

    @property
    def documentsPath(self):
        """Absolute path of the `documents` directory."""
        return join(self.directory, self.documents)

    @property
    def bibfilePath(self):
        """Absolute path of the bib file."""
        return join(self.directory, self.bibfileName)

    @property
    def configPath(self):
        """Absolute path of the conf file."""
        return join(self.directory, BTVCSCONF)

    def referencedDocuments(self):
        docs = []
        for entry in self.bibfile.values():
            fname = entry.filename()
            if fname:
                docs.append(fname)
        return docs

    def existingDocuments(self):
        """Walks recursively through the :attr:`documents` directory and return the paths of all files
        contained in there, relative to :attr:`documentsPath`.
        """
        for dirpath, _, filenames in os.walk(self.documentsPath):
            for file in filenames:
                yield relpath(join(dirpath, file), self.documentsPath)

    def strval(self, value):
        if isinstance(value, MacroReference):
            if value.name in self.bibfile.macroDefinitions:
                return self.bibfile.macroDefinitions[value.macro]
            if value.name in self.journals:
                return self.journals[value.name].full
        return str(value)

    def makeJournalBibfiles(self):
        """Creates or updates the files containing journal macro definitions in full and
        abbreviated form, respectively.
        """
        base = self.bibfilePath[:-4]
        if not exists(base + '_abbr.bib') or not exists(base + '_full.bib') or \
                os.path.getmtime(self.bibfilePath) > os.path.getmtime(base + '_abbr.bib'):
            self.journals.writeBibfiles(base)

    def runJabref(self):
        """Tries to open this database's ``.bib`` file with `JabRef`_. Will do the following:

        - If there is a file named ``jabref.jar`` in :attr:`directory`, it is run with the
          java interpreter through ``java -jar jabref.jar``. If additionally there is a file
          ``jabref.prefs``, JabRef`s options are imported from that file.
        - Otherwise, the command ``jabref`` is executed. To that end, the ``jabref`` binary must
          be in the system's ``PATH``.

        :returns: The :class:`subprocess.Popen` object corresponding to JabRef process.
        """
        shell = False
        if exists(join(self.directory, 'jabref.jar')):
            if os.name == 'nt':
                cmdline = ['start', 'jabref.jar']
                shell = True
            else:
                cmdline = ['java', '-jar', 'jabref.jar']
        else:
            cmdline = ['jabref']
        if exists(join(self.directory, 'jabref.prefs')):
            cmdline += ['--primp', join('jabref.prefs')]
        cmdline.append(os.curdir + os.sep + self.bibfileName)
        try:
            return subprocess.Popen(cmdline, shell=shell, cwd=self.directory)
        except FileNotFoundError as fnf:
            if cmdline[0] in ('java', 'start'):
                fnf.strerror = 'Please install Java from http://java.com.'
            elif cmdline[0] == 'jabref':
                fnf.strerror = 'Please install JabRef from http://jabref.sf.net.'
            raise fnf

    @property
    def vcs(self):
        """The :class:`VCSInterface` object associated to this db.

        Will be created on first access.
        """
        if self._vcs is None:
            self._vcs = VCSInterface.get(self)
        return self._vcs

    def export(self, templateString=None, docDir=None):
        """Exports the BibTeX database to a string by using the jinja template engine."""
        import datetime, hashlib
        try:
            import jinja2
        except ImportError:
            raise ImportError('You need to install the jinja2 package in order to export.')
        if docDir is None:
            docDir = self.documentsPath
        def md5filter(value):
            return hashlib.md5(value.encode()).hexdigest()
        env = jinja2.Environment(autoescape=False)
        env.filters['md5'] = md5filter
        if templateString is None:
            from pkg_resources import resource_string
            templateString = resource_string(__name__, 'defaultTemplate.html').decode('UTF-8')
        template = env.from_string(templateString)
        revision = self.vcs.revision()
        now = datetime.datetime.now().strftime('%c')
        return template.render(database=self, docDir=docDir, revision=revision, now=now)



class Journal:
    """A single journal entry in the journals file.

    .. attribute:: macro

        The macro defining the journal.
    .. attribute:: abbr

        The abbreviated journal name.
    .. attribute:: full

        The full (unabbreviated) journal name.
    """
    def __init__(self, macro, abbr, full):
        self.macro = macro
        self.abbr = abbr
        self.full = full

    def __iter__(self):
        yield self.full
        yield self.abbr
        yield self.macro


class JournalsFile(OrderedDict):
    """Represents the file in a bibtexvcs database containing journal abbreviations.

    In the journal file, each journal is represented by a configuration section of the
    following form::

        [JOURNAL_MACRO]
        full = A Journal of Nice and Interesting Topics
        abbr = J. Nice Int. Top.

    :param filename: The path of the journals file.
    :type  filename: str
    :param journals: As an alternative of giving a ``filename``, you can also provide a list
                     of journals entries to create the :class:`JournalsFile` object.
    :type journals: iterable
    """

    def __init__(self, filename=None, journals=None):
        super(JournalsFile, self).__init__()

        if filename:
            journals = []
            parser = configparser.ConfigParser()
            try:
                parser.read(filename, encoding='UTF-8')
            except configparser.Error as error:
                raise DatabaseFormatError('Malformed journals file:\n\n{}'.format(error))
            for section in parser.sections():
                journals.append(Journal(macro=section,
                                        abbr=parser.get(section, "abbr"),
                                        full=parser.get(section, "full")))
        for journal in sorted(journals, key=lambda j: j.full):
            self[journal.macro] = journal

    def write(self, filename):
        """Write the journals file to the given filename."""
        config = configparser.ConfigParser()
        for journal in self.values():
            config.add_section(journal.macro)
            config[journal.macro]['abbr'] = journal.abbr
            config[journal.macro]['full'] = journal.full
        with io.open(filename, 'w', encoding='UTF-8', newline='\n') as journalfile:
            config.write(journalfile)

    def writeBibfiles(self, basename):
        """Creates two ``.bib`` files that contain macro definitions for all journals defined in
        this file, resolving to the full and abbreviated journal names, respectively.

        `basename` is the base path of the output files, which will have the file names
        ``<basename>_full.bib`` and ``<basename>_abbr.bib``, respectively.
        """
        for jrnlType in 'full', 'abbr':
            outFile = '{}_{}.bib'.format(basename, jrnlType)
            with io.open(outFile, 'w', encoding='UTF-8') as bibfile:
                for journal in self.values():
                    bibfile.write('@STRING{' + journal.macro + ' = {' + getattr(journal, jrnlType)
                                  + '}}\n')



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import argparse, io

from bibtexvcs.database import Database
from bibtexvcs import config

def export(args):
    if args.template:
        with io.open(args.template, 'r', encoding='UTF-8') as templateFile:
            templateString = templateFile.read()
    else:
        templateString = None
    output = args.db.export(templateString=templateString, docDir=args.docs)
    with io.open(args.output, 'wt', encoding='UTF-8') as outfile:
        outfile.write(output)

def script():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--database', metavar='DB', help='database root directory')
    subparsers = parser.add_subparsers(title='Commands')
    parser_export = subparsers.add_parser('export', help='export database using a template')
    parser_export.add_argument('--template', help='template file')
    parser_export.add_argument('--docs', help='documents root path')
    parser_export.add_argument('output', help='output file')
    parser_export.set_defaults(func=export)
    parser_check = subparsers.add_parser('check', help='check database consistency')

    args = parser.parse_args()
    if args.database:
        args.db = Database(args.database)
    else:
        args.db = config.getDefaultDatabase()
    args.func(args)

if __name__ == '__main__':
    script()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

# The parser definitions are partially based on "btpyparse" by Matthew Brett
# 2010 Simplified BSD license.
# https://github.com/matthew-brett/btpyparse

from __future__ import division, print_function, unicode_literals
import re

from pyparsing import (Regex, Suppress, ZeroOrMore, OneOrMore, Group, Optional, Forward,
                       SkipTo, CaselessLiteral, Dict, originalTextFor, delimitedList)

from bibtexvcs.bibfile import (Entry, Comment, ImplicitComment, MacroReference,
                                MacroDefinition, Name, Preamble)

LCURLY = Suppress('{')
RCURLY = Suppress('}')
QUOTE = Suppress('"')
COMMA = Suppress(',')
AT = Suppress('@')
EQUALS = Suppress('=')
HASH = Suppress('#')

bracketed = lambda expr: LCURLY + expr + RCURLY

# Define parser components for strings (the hard bit)
charsNoCurly = Regex(r"[^{}]+")
charsNoCurly.leaveWhitespace()
charsNoQuotecurly = Regex(r'[^"{}]+')
charsNoQuotecurly.leaveWhitespace()

# Curly string is some stuff without curlies, or nested curly sequences
curlyString = Forward().leaveWhitespace()
curlyItem = Group(curlyString) | charsNoCurly
curlyString <<= LCURLY + ZeroOrMore(curlyItem) + RCURLY

# quoted string is either just stuff within quotes, or stuff within quotes, within
# which there is nested curliness
quotedItem = Group(curlyString) | charsNoQuotecurly
quotedString = QUOTE + ZeroOrMore(quotedItem) + QUOTE

number = Regex('[0-9]+')
# Basis characters (by exclusion) for variable / field names.  The following
# list of characters is from the btparse documentation
anyName = Regex('[^\s"#%\'(),={}]+')

# btparse says, and the test bibs show by experiment, that macro and field names
# cannot start with a digit.  In fact entry type names cannot start with a digit
# either (see tests/bibs). Cite keys can start with a digit
notDigname = Regex('[^\d\s"#%\'(),={}][^\s"#%\'(),={}]*')

comment = AT + CaselessLiteral('comment') + LCURLY + charsNoCurly.setResultsName("comment") + RCURLY
comment.setParseAction(Comment.fromParseResult)

# The name types with their digiteyness
notDigLower = notDigname.copy().setParseAction(lambda t: t[0].lower())

macroDef = notDigLower.copy()

macroRef = notDigLower.copy().setParseAction(MacroReference.fromParseResult)
fieldName = notDigLower.copy()
entryType = notDigLower.setResultsName('entry type')
citeKey = anyName.setResultsName('cite key')
string = (number | macroRef | quotedString | curlyString)

# There can be hash concatenation
fieldValue = string + ZeroOrMore(HASH + string)

namePart = Regex(r'(?!\band\b)\w+\.?') | curlyString
nobility = Regex(r'[a-z]\w+\.?(\s[a-z]\w+\.?)*').setResultsName("nobility")  # "van" etc.
spacedNames = originalTextFor(OneOrMore(namePart))
firstNames = spacedNames.copy().setResultsName("firstname")
lastNames = spacedNames.copy().setResultsName("lastname")
nameSuffix = namePart.copy().setResultsName("suffix")

# a name in "comma separated" style, like "Helmling, Michael"
csName = (Optional(nobility) + lastNames + COMMA + Optional(nameSuffix + COMMA) + firstNames)


def labelLiteralName(toks):
    """In case of a literal name, we cannot distinguish between first and middle names, or
    recognize multi-part last names. Hence it is assumed that the last part is the last name,
    anything else is stored as first names.
    """
    toks["lastname"] = toks[-1]
    if len(toks) > 1:
        toks["firstname"] = " ".join(toks[:-1])
    return toks

# a name in "literal" style, like "Michael Helmling
literalName = OneOrMore(namePart).setParseAction(labelLiteralName)

def makeName(toks):
    """Create a :class:`Name` object from the parse result of either a csName or a literalName."""
    return Name(first=toks.get("firstname"),
                nobility=toks.get("nobility"),
                last=toks.get("lastname"),
                suffix=toks.get("suffix"))

name = (csName | literalName).setParseAction(makeName)
NAME_SEP = Regex(r'and[^}]', flags=re.IGNORECASE).suppress()
namesList = LCURLY + delimitedList(name, NAME_SEP) + RCURLY

author = CaselessLiteral("author")
# we treat the author field special because we parse names
fieldDef = Group((author + EQUALS + namesList)) | Group((fieldName + EQUALS + fieldValue))
entryContents = Dict(ZeroOrMore(fieldDef + COMMA) + Optional(fieldDef))


entry = originalTextFor(AT + entryType + bracketed(citeKey + COMMA + entryContents), asString=False)
entry.addParseAction(Entry.fromParseResult)

# Preamble is a macro-like thing with no name
preamble = AT + CaselessLiteral('preamble') + bracketed(fieldValue).setResultsName("preamble")
preamble.setParseAction(Preamble.fromParseResult)

# Macros (aka strings)
macroContents = macroDef.setResultsName("macro") + EQUALS + fieldValue.setResultsName("definition")


macro = (AT + CaselessLiteral('string') + bracketed(macroContents)).setParseAction(MacroDefinition.fromParseResult)

icomment = SkipTo('@').setResultsName("comment").setParseAction(ImplicitComment.fromParseResult)

definitions = comment | preamble | macro | entry
bibfile = Optional(icomment) + ZeroOrMore(definitions)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import concurrent.futures, tempfile, webbrowser, sys
import os.path
from contextlib import contextmanager

# we support PyQt5, PyQt4 and PySide
PYSIDE = False
try:
    from PyQt5 import QtWidgets, QtCore
    from PyQt5.QtCore import Qt
    QT5 = True
except ImportError:
    QT5 = False
    try:
        from PyQt4 import QtGui, QtCore
        from PyQt4.QtCore import Qt
        QtWidgets = QtGui
    except ImportError:
        from PySide import QtGui, QtCore
        from PySide.QtCore import Qt
        QtWidgets = QtGui
        PYSIDE = True

from bibtexvcs.vcs import MergeConflict, typeMap, AuthError, VCSInterface
from bibtexvcs.database import Database, Journal, JournalsFile, DatabaseFormatError, BTVCSCONF
from bibtexvcs import config

def standardIcon(widget, standardPixmap):
    """Helper function to generate a QIcon from a standardPixmap."""
    return widget.style().standardIcon(getattr(widget.style(), standardPixmap))

class BtVCSGui(QtWidgets.QWidget):
    """Main window of the BibTeX VCS GUI application.
    """

    def __init__(self):
        super(BtVCSGui, self).__init__()
        self.setWindowTitle("BibTeX VCS")

        layout = QtWidgets.QVBoxLayout()

        dbLayout = QtWidgets.QHBoxLayout()
        self.dbLabel = QtWidgets.QLabel("Please open a database")
        dbSelect = QtWidgets.QPushButton(standardIcon(self, "SP_DialogOpenButton"), "Open")
        dbSelect.setToolTip("Open existing local checkout of a BibTeX VCS database")
        dbSelect.clicked.connect(self.openDialog)
        dbClone = QtWidgets.QPushButton(standardIcon(self, "SP_DriveNetIcon"), "Clone")
        dbClone.setToolTip("Clone a BibTeX VCS database from a remote repository")
        dbClone.clicked.connect(self.cloneDialog)

        dbLayout.addWidget(self.dbLabel)
        dbLayout.addStretch()
        dbLayout.addWidget(dbSelect)
        dbLayout.addWidget(dbClone)
        layout.addLayout(dbLayout)
        self.linkLabel = QtWidgets.QLabel('')
        self.linkLabel.setOpenExternalLinks(True)
        layout.addWidget(self.linkLabel)
        self.setLayout(layout)
        self.guiComplete = False

        # helpers for asynchronous function calls
        self.executor = concurrent.futures.ThreadPoolExecutor(1)
        self.futureTimer = QtCore.QTimer(self)
        self.futureTimer.setInterval(10)
        self.futureTimer.timeout.connect(self._updateFuture)
        self.prog = QtWidgets.QProgressDialog(self)
        self.prog.setRange(0, 0)
        if QT5 or PYSIDE:
            self.prog.setCancelButtonText(None)
        else:
            self.prog.setCancelButtonText(QtCore.QString())
        self.prog.setWindowModality(Qt.WindowModal)

        self.db = None
        self.show()

        self.runAsync("Opening previous database ...", self.loadDefault, config.getDefaultDatabase)

    def loadDefault(self):
        """Opens the default database as defined in :mod:`bibtexvcs`'s config file. If no default
        database is configured, nothing happens.
        """
        lastDB = self.future.result()
        if lastDB:
            self.setDB(lastDB)

    def setDB(self, db):
        """Set the current database to `db`."""
        self.db = db
        if not self.guiComplete:
            buttonLayout = QtWidgets.QHBoxLayout()
            updateButton = QtWidgets.QPushButton(standardIcon(self, "SP_ArrowDown"), "1: Update")
            updateButton.clicked.connect(self.updateRepository)
            jabrefButton = QtWidgets.QPushButton("2: JabRef")
            jabrefButton.clicked.connect(self.jabref)
            commitButton = QtWidgets.QPushButton(standardIcon(self, "SP_ArrowUp"), "3: Commit")
            commitButton.clicked.connect(self.runChecks)
            for button in updateButton, jabrefButton, commitButton:
                buttonLayout.addWidget(button)
            self.journalsTable = JournalsWidget(self.db)
            self.layout().addWidget(self.journalsTable)
            self.layout().addLayout(buttonLayout)
            self.guiComplete = True
        self.reload()
        self.updateRepository()

    def runAsync(self, labelText, finishedCall, fn, *args, **kwargs):
        """Helper function for asynchronous calls during which a progress dialog is shown. The
        progress dialog will be labelled by `labelText`.

        `finishedCall` is an optional callable that will be called after `fn` has finished running.

        `fn` is the callable to be run asynchronously. Additional arguments are passed to this
        function.
        """
        self.prog.setLabelText(labelText)
        self.prog.show()
        self.finishedCall = finishedCall
        self.future = self.executor.submit(fn, *args, **kwargs)
        self.futureTimer.start()

    def _updateFuture(self):
        if self.future.done():
            self.futureTimer.stop()
            self.prog.close()
            if self.finishedCall:
                self.finishedCall()

    @contextmanager
    def catchExceptions(self, onAuthEntered=None):
        """Contextmanager catching common exceptions and displaying message boxes if they occur.

        The caught exceptions are :class:`DatabaseFormatError`, :class:`MergeConflict` and
        :class:`AuthError`.

        `onAuthEntered` is an optional callable. If it is provided and an :class:`AuthError` is
        caught, the following happens:
        - A :class:`LoginDialog` is shown to ask the user for a login,
        - if the dialog is canceled, the :class:`AuthError` is re-raised
        - otherwise, the function `onAuthEntered` is called.

        If the `store login` checkbox in the :class:`LoginDialog` was checked, the information is
        set to `self.db.vcs` and :fun:`config.setAuthInformation` is called.
        """
        M = QtWidgets.QMessageBox
        try:
            yield
        except DatabaseFormatError as e:
            M.critical(self, 'Error Opening Database', str(e))
        except MergeConflict as mc:
            M.critical(self, "Merge conflict", str(mc))
        except AuthError as a:
            if onAuthEntered is not None:
                ans = LoginDialog.getLogin(self, str(a))
                if ans:
                    self.db.vcs.username, self.db.vcs.password, store = ans
                    if store:
                        config.setAuthInformation(self.db)
                    onAuthEntered()
                    return
            M.critical(self, "Authorization Required", str(a))
        except Exception as e:
            M.critical(self, 'Unhandled exception', str(e))

    def reload(self):
        self.journalsTable.setDB(self.db)
        if self.db.publicLink:
            self.linkLabel.setText('Web: <a href="{0}">{0}</a>'.format(self.db.publicLink))
        self.linkLabel.setVisible(self.db.publicLink is not None)
        self.dbLabel.setText("Database: <i>{}</i>".format(self.db.directory))
        self.setWindowTitle("BibTeX VCS: {}".format(self.db.name))

    def openDialog(self):
        ans = QtWidgets.QFileDialog.getOpenFileName(self, "Select Database", "",
                    "BibTeX VCS configuration files ({})".format(BTVCSCONF))
        filename = ans[0] if QT5 or PYSIDE else ans
        if len(filename) > 0:
            directory = os.path.dirname(str(filename))
            with self.catchExceptions():
                db = Database(directory)
                self.setDB(db)

    def cloneDialog(self):
        self.cloneDialog = CloneDialog(self)
        if self.cloneDialog.exec_() == self.cloneDialog.Accepted:
            self.login = dict()
            self.storeLogin = False
            self.cloneDB_init()

    def cloneDB_init(self):
        self.runAsync("Cloning database ... ",
                      self.cloneDB_handle,
                      VCSInterface.getClonedDatabase,
                      self.cloneDialog.urlEdit.text(),
                      self.cloneDialog.targetEdit.text(),
                      self.cloneDialog.vcsTypeChooser.currentText(),
                      **self.login)

    def cloneDB_handle(self):
        with self.catchExceptions():
            try:
                self.setDB(self.future.result())
                if self.storeLogin:
                    config.setAuthInformation(self.db)
            except AuthError as e:
                ans = LoginDialog.getLogin(self, str(e))
                if ans:
                    self.login['username'], self.login['password'], self.storeLogin = ans
                    self.cloneDB_init()
                else:
                    raise e

    def updateRepository(self):
        self.runAsync("Updating repository ...", self.update_handle, self.db.vcs.update)

    def update_handle(self):
        with self.catchExceptions(onAuthEntered=self.updateRepository):
            changed = self.future.result()
            if changed:
                QtWidgets.QMessageBox.information(self,
                        "Update successful", "Successfully merged remote changes")
                self.reload()

    def jabref(self):
        try:
            self.db.runJabref()
        except FileNotFoundError as e:
            QtWidgets.QMessageBox.critical(self, 'Could not start JabRef', str(e))

    def runChecks(self):
        self.runAsync("Performing database checks ...", self.runChecks_handle, self.runChecks_init)

    def runChecks_init(self):
        from bibtexvcs import checks
        self.db.reload()
        return checks.performDatabaseCheck(self.db)

    def runChecks_handle(self):
        self.reload()
        ans = self.future.result()
        if len(ans) > 0:
            title = "Database Check Failed"
            text = 'One or more database checks failed. Please fix, then try again.'
            detailed = "\n\n".join(str(a) for a in ans)
            box = QtWidgets.QMessageBox(self)
            box.setWindowTitle(title)
            box.setText(text)
            box.setDetailedText(detailed)
            box.setStandardButtons(box.Close)
            box.exec_()
        else:
            self.commit_init()

    def commit_init(self):
        self.runAsync("Committing repository ...", self.commit_handle, self.db.vcs.commit)

    def commit_handle(self):
        with self.catchExceptions(onAuthEntered=self.commit_init):
            if not self.future.result():
                QtWidgets.QMessageBox.information(self, "No Local Changes", "Nothing to commit.")
            else:
                QtWidgets.QMessageBox.information(self, "Commit successful", "Commit successful.")
                self.reload()

    def closeEvent(self, event):
        if self.db and self.db.vcs.localChanges():
            ans = QtWidgets.QMessageBox.question(self, "Local changes present",
                    "Database was modified locally. Are you sure you want to quit "
                    "without committing the changes?",
                    QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Yes)
            if ans == QtWidgets.QMessageBox.Yes:
                config.setDefaultDatabase(self.db)
                event.accept()
            else:
                event.ignore()
        else:
            if self.db:
                config.setDefaultDatabase(self.db)
            event.accept()


class JournalsWidget(QtWidgets.QWidget):

    def __init__(self, db):
        super(JournalsWidget, self).__init__()
        self.table = QtWidgets.QTableWidget(len(db.journals), 3)
        self.table.setHorizontalHeaderLabels(["Full", "Abbreviated", "Macro"])
        self.setDB(db)
        try:  # Qt5
            self.table.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
            self.table.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
            self.table.horizontalHeader().setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeToContents)
            self.table.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        except AttributeError:  # Qt4
            self.table.horizontalHeader().setResizeMode(0, QtWidgets.QHeaderView.Stretch)
            self.table.horizontalHeader().setResizeMode(1, QtWidgets.QHeaderView.Stretch)
            self.table.horizontalHeader().setResizeMode(2, QtWidgets.QHeaderView.ResizeToContents)
            self.table.verticalHeader().setResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        layout = QtWidgets.QVBoxLayout()
        journalsLabel = QtWidgets.QLabel("<h3>Journals Management</h3>")
        newJournalButton = QtWidgets.QPushButton("Add Journal")
        delJournalButton = QtWidgets.QPushButton(standardIcon(self, "SP_TrashIcon"), "Delete")
        delJournalButton.clicked.connect(self.deleteCurrent)
        newJournalButton.clicked.connect(self.addJournal)

        buttonLayout = QtWidgets.QHBoxLayout()
        buttonLayout.addWidget(journalsLabel)
        buttonLayout.addStretch()
        buttonLayout.addWidget(newJournalButton)
        buttonLayout.addWidget(delJournalButton)
        layout.addLayout(buttonLayout)
        layout.addWidget(self.table)
        self.setLayout(layout)
        self.table.cellChanged.connect(self.updateJournalsFile)
        self.dontUpdate = False
        self.setContentsMargins(0, 0, 0, 0)
        layout.setContentsMargins(0, 0, 0, 0)

    def makeItem(self, text, editable=True):
        item = QtWidgets.QTableWidgetItem(text)
        if not editable:
            item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        return item

    def setDB(self, db):
        self.db = db
        self.dontUpdate = True
        self.table.clearContents()
        self.table.setRowCount(len(self.db.journals))
        for i, journal in enumerate(self.db.journals.values()):
            self.table.setItem(i, 0, self.makeItem(journal.full))
            self.table.setItem(i, 1, self.makeItem(journal.abbr))
            self.table.setItem(i, 2, self.makeItem(journal.macro, False))
        self.dontUpdate = False

    def deleteCurrent(self):
        currentRow = self.table.currentRow()
        self.table.removeRow(currentRow)
        self.updateJournalsFile()

    def addJournal(self):
        macro, ok = QtWidgets.QInputDialog.getText(self.parent(), "New Journal's Macro",
            "Please enter the <i>Macro</i> of the new journal")
        if not ok:
            return
        for i in range(self.table.rowCount()):
            if self.table.item(i, 2).text() == macro:
                QtWidgets.QMessageBox.critcical(self.parent(), "Macro exists",
                    "The macro '{}' is already chosen by another journal.".format(macro))
                return
        index = self.table.rowCount()
        self.dontUpdate = True
        self.table.insertRow(index)
        self.table.setItem(index, 2, self.makeItem(macro, False))
        # initialize abbr and full with the macro
        self.table.setItem(index, 0, self.makeItem(macro))
        self.dontUpdate = False
        self.table.setItem(index, 1, self.makeItem(macro))

    def updateJournalsFile(self):
        if self.dontUpdate:
            return
        journals = []
        for i in range(self.table.rowCount()):
            journals.append(Journal(full=self.table.item(i, 0).text(),
                                    abbr=self.table.item(i, 1).text(),
                                    macro=self.table.item(i, 2).text()))
        self.db.journals = JournalsFile(journals=journals)
        self.db.journals.write(self.db.journalsPath)


class CloneDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super(CloneDialog, self).__init__(parent)
        layout = QtWidgets.QGridLayout()
        self.urlEdit = QtWidgets.QLineEdit()
        self.urlEdit.setPlaceholderText("remote repository URL")
        layout.addWidget(QtWidgets.QLabel("URL:"), 0, 0)
        layout.addWidget(self.urlEdit, 0, 1)
        self.vcsTypeChooser = QtWidgets.QComboBox()
        self.vcsTypeChooser.addItems([t for t in typeMap if t is not None])
        layout.addWidget(self.vcsTypeChooser, 0, 2)
        layout.addWidget(QtWidgets.QLabel('Target:'), 1, 0)
        self.targetEdit = QtWidgets.QLineEdit()
        self.targetEdit.setPlaceholderText('local checkout directory')
        layout.addWidget(self.targetEdit, 1, 1)
        chooseTargetButton = QtWidgets.QPushButton('Choose...')
        chooseTargetButton.clicked.connect(self.chooseTarget)
        layout.addWidget(chooseTargetButton, 1, 2)
        Ok = QtWidgets.QDialogButtonBox.Ok
        Cancel = QtWidgets.QDialogButtonBox.Cancel
        self.btbx = QtWidgets.QDialogButtonBox(Cancel | Ok)
        self.btbx.button(Ok).setEnabled(False)
        layout.addWidget(self.btbx, 2, 0, 1, 3)
        layout.setColumnMinimumWidth(1, 250)
        self.btbx.rejected.connect(self.reject)
        self.btbx.accepted.connect(self.accept)
        self.targetEdit.textChanged.connect(self.checkOk)
        self.urlEdit.textChanged.connect(self.checkOk)
        self.setLayout(layout)

    def checkOk(self):
        ok = all((edit.text().strip() != '' for edit in (self.targetEdit, self.urlEdit)))
        self.btbx.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(ok)

    def chooseTarget(self):
        target = QtWidgets.QFileDialog.getExistingDirectory(self, 'Choose Target')
        if target:
            self.targetEdit.setText(target)


class LoginDialog(QtWidgets.QDialog):

    def __init__(self, message=None, parent=None):
        super(LoginDialog, self).__init__(parent)

        layout = QtWidgets.QGridLayout()
        if message:
            layout.addWidget(QtWidgets.QLabel(message), 0, 0, 1, 2)
        layout.addWidget(QtWidgets.QLabel('User:'), 1, 0)
        layout.addWidget(QtWidgets.QLabel('Password:'), 2, 0)
        self.userEdit = QtWidgets.QLineEdit()
        self.userEdit.setPlaceholderText('enter username')
        layout.addWidget(self.userEdit, 1, 1)
        self.passEdit = QtWidgets.QLineEdit()
        self.passEdit.setPlaceholderText('enter password')
        self.passEdit.setEchoMode(QtWidgets.QLineEdit.Password)
        layout.addWidget(self.passEdit, 2, 1)
        self.storeBox = QtWidgets.QCheckBox('Remember login')
        layout.addWidget(self.storeBox, 3, 1)
        btbx = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Cancel |
                                          QtWidgets.QDialogButtonBox.Ok)
        btbx.accepted.connect(self.accept)
        btbx.rejected.connect(self.reject)
        layout.addWidget(btbx, 4, 0, 1, 2)

        self.setLayout(layout)

    @staticmethod
    def getLogin(parent=None, message=None):
        dialog = LoginDialog(parent=parent, message=message)
        if dialog.exec_() == dialog.Accepted:
            return dialog.userEdit.text(), dialog.passEdit.text(), dialog.storeBox.isChecked()
        return None


def run():
    import bibtexvcs, pkg_resources
    from distutils.version import StrictVersion
    app = QtWidgets.QApplication(sys.argv)
    window = None
    newVersion = bibtexvcs.pypiVersion()
    if newVersion:
        newVersion = StrictVersion(newVersion)
        myVersion = StrictVersion(pkg_resources.get_distribution('bibtexvcs').version)
        if newVersion > myVersion:
            window = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Critical,
                                           "New version available",
                                           "A new version of BibTeX VCS ({}) is available. "
                                           "Please update, then start again.".format(newVersion),
                                           QtWidgets.QMessageBox.Ok)
            window.show()
            window.accepted.connect(app.exit)
    if window is None:
        window = BtVCSGui()
    app.exec_()

if __name__ == '__main__':
    run()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import unittest

from bibtexvcs import bibfile, parser

bibtext = """This is an implicit comment.
@PREAMBLE{"This is the preamble."}

@ArtiCLe{ArticleKey,
    author = {Helmling, Michael J.},
    title  = {A bibtex database under revision control}
}
"""
class TestBasicParsing(unittest.TestCase):

    def setUp(self):
        self.bibfile = bibfile.BibFile(bibstring=bibtext)

    def testImplicitComment(self):
        self.assertGreater(len(self.bibfile.comments), 0)
        self.assertIsInstance(self.bibfile.comments[0], bibfile.ImplicitComment)
        self.assertEqual(self.bibfile.comments[0].comment,
                         "This is an implicit comment.\n")

    def testEntry(self):
        self.assertGreater(len(self.bibfile), 0)
        self.assertIn('ArticleKey', self.bibfile, 'ArticleKey not found in {}'.format(list(self.bibfile.keys())))
        entry = self.bibfile['ArticleKey']
        self.assertEqual(entry.entrytype, 'article')


class TestNameParsing(unittest.TestCase):

    def parse(self, name):
        return parser.name.parseString(name, parseAll=True)[0]

    def testNobility(self):
        name = "van der Zalm, E."
        parsed = self.parse(name)
        self.assertEqual(parsed.nobility, "van der")

        name = "van Emde Boas, P."
        parsed = self.parse(name)
        self.assertEqual(parsed.last, "Emde Boas")
        self.assertEqual(parsed.nobility, "van")
        self.assertEqual(parsed.first, "P.")

    def testCommaName(self):
        parsed = self.parse("Helmling, Michael")
        self.assertEqual(parsed.first, "Michael")
        self.assertEqual(parsed.last, "Helmling")
        self.assertIsNone(parsed.nobility)
        self.assertIsNone(parsed.suffix)

        parsed = self.parse("Forney, Jr., David G.")
        self.assertEqual(parsed.first, "David G.")
        self.assertEqual(parsed.suffix, "Jr.")
        self.assertEqual(parsed.last, "Forney")

    def testLiteralName(self):
        parsed = self.parse("Michael Jakob Helmling")
        self.assertEqual(parsed.last, "Helmling")
        self.assertEqual(parsed.first, "Michael Jakob")

    def testCurlyName(self):
        parsed = self.parse("{Ministry of Truth and Justice}")
        self.assertEqual(parsed.last, "Ministry of Truth and Justice")


class TestNamesListParsing(unittest.TestCase):

    def parse(self, namesList):
        return parser.namesList.parseString(namesList, parseAll=True)

    def testLiteralAnd(self):
        parsed = self.parse("{{Ministry of Truth and Justice} and Helmling, Michael}")
        self.assertEqual(len(parsed), 2)
        ministry = parsed[0]
        self.assertEqual(ministry.last, "Ministry of Truth and Justice")
        me = parsed[1]
        self.assertEqual(me.last, "Helmling")



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import unittest
from bibtexvcs import bibfile

class TestDatabaseChecks():
    pass


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import io, os, subprocess, sys
from bibtexvcs import config

class MergeConflict(Exception):
    """Raised when merge conflicts occur on a VCS update operation."""
    pass

class AuthError(Exception):
    """Raised on authentication errors when accessing remote repositories."""
    pass

typeMap = {}

class VCSMeta(type):
    def __init__(cls, name, base, namespace):
        super(VCSMeta, cls).__init__(name, base, namespace)
        if hasattr(cls, 'vcsType'):
            typeMap[cls.vcsType] = cls

VCSBase = VCSMeta('VCSBase' if sys.version_info.major >= 3 else b'VCSBase', (object,), {})

class VCSInterface(VCSBase):
    """Interface to the version control system (VCS) of a :mod:`bibtexvcs` database.

    .. attribute:: username

        Optional username for authentication.
    .. attribute:: password

        Optional password for authentication.
    """

    def __init__(self, db):
        self.db = db
        ans = config.getAuthInformation(db)
        if ans:
            self.username, self.password = ans
        else:
            self.username = self.password = None

    @property
    def root(self):
        return self.db.directory

    def add(self, path):
        """Adds a file or directory to the repository.

        :param path: The path (relative to the vcs root).
        :type path: str
        """

    def localChanges(self):
        """Determines whether any local changes have been made to the repository.

        Returns ``True`` if one of the following requirements are met:

        - any versioned file has been modified,
        - any versioned file in the documents folder has been deleted, or
        - any unversioned file has been placed in the documents folder.
        """
        raise NotImplementedError()

    def update(self):
        """Checks if there are updates in the remote repository. If so, tries to merge
        them and reloads the database.

        :return: ``True`` if something was update, ``False`` otherwise.
        :raises: :class:`MergeConflict` if a conflict is introduced by merging."""
        raise NotImplementedError()

    def commit(self, msg=None):
        """Commit local changes and push to remote, if exists.

        :param msg: Commit message.
        :type msg:  str

        :returns: Indicate whether there were any actual changes.
        :rtype  : bool"""
        raise NotImplementedError()

    def revision(self):
        """Returns a tuple containing the current revision and its date."""
        raise NotImplementedError()

    @staticmethod
    def get(db):
        try:
            return typeMap[db.vcsType](db)
        except KeyError:
            raise KeyError("No VCS type '{}' is known".format(db.vcsType))

    @classmethod
    def clone(cls, url, target):
        """Clones a remote repository.

        :param url: The remote repository URL.
        :type url: str
        :param target: The target directory.
        :type target: str
        """
        raise NotImplementedError()

    @staticmethod
    def getClonedDatabase(url, target, vcsType, username=None, password=None):
        vcsCls = typeMap[vcsType]
        vcsCls.clone(url, target, username=username, password=password)
        from bibtexvcs.database import Database
        return Database(target)


class MercurialInterface(VCSInterface):
    """Interface to the `Mercurial <http://mercurial.selenic.com>`_ version control system.
    """
    vcsType = 'mercurial'
    cmdline = ['hg', '--noninteractive', '--config', 'auth.x.prefix=*']

    def __init__(self, *args, **kwargs):
        super(MercurialInterface, self).__init__(*args, **kwargs)
        self.hasRemote = len(self.callHg("showconfig", "paths.default")) > 0

    def callHg(self, *args):
        """Calls the ``hg`` script in the database directory with the given arguments.
        """
        return MercurialInterface._callHg(*args, username=self.username, password=self.password,
                                          cwd=self.root)

    @classmethod
    def _callHg(cls, *args, **kwargs):
        cmdline = cls.cmdline[:]
        kwargs = kwargs.copy()
        if kwargs.get('username'):
            cmdline += ['--config', 'auth.x.username={}'.format(kwargs['username'])]
        if kwargs.get('password'):
            cmdline += ['--config', 'auth.x.password={}'.format(kwargs['password'])]
        if 'username' in kwargs:
            del kwargs['username']
        if 'password' in kwargs:
            del kwargs['password']
        env = os.environ.copy()
        env['LANG'] = 'C'
        env['LANGUAGE'] = 'C'  # on Ubuntu 12.04, LANG=C does not convince hg to use english output
        try:
            return subprocess.check_output(cmdline + list(args), env=env,
                                           stderr=subprocess.STDOUT, **kwargs)
        except subprocess.CalledProcessError as e:
            output = e.output.decode(errors='replace')
            if "authorization required" in output:
                raise AuthError('Authorization required for the mercurial repository.')
            if "authorization failed" in output:
                raise AuthError('Authorization for the mercurial repository failed.')
            if 'conflicts during merge' in output:
                raise MergeConflict('Conflict arised when merging remote and local changes!\n'
                                    'You need to fix this issue by hand. Error message:\n{}'
                                    .format(e.output))
            if 'unresolved merge conflicts' in output:
                raise MergeConflict('There are unresolved merge conflicts in your repository.\n'
                                    'You have to fix them manually before proceeding to use this '
                                    'tool.')
            raise e

    def add(self, path):
        self.callHg('add', path)

    def localChanges(self):
        if len(self.callHg('status', '--modified', '--added')) > 0:
            return True
        # by this we ignore unversioned files in the base folder
        return len(self.callHg('status', '--deleted', '--unknown', self.db.documents)) > 0

    def update(self):
        if self.hasRemote:
            cmd = ['pull', '--update']
        else:
            cmd = ['update']
        ans = self.callHg(*cmd).decode()
        if 'no changes found' in ans:
            return False
        self.db.reload()
        return True

    def commit(self, msg=None):
        if not self.localChanges():
            return False
        if os.name != 'nt':
            with io.open(self.db.bibfilePath, 'rt', encoding='UTF-8') as bibfile:
                bib = bibfile.read()
            with io.open(self.db.bibfilePath, 'wt', encoding='UTF-8', newline='\r\n') as bibfile:
                bibfile.write(bib)
        hgOutput = self.callHg('status', '--deleted', '--no-status', self.db.documents)
        deletedDocs = hgOutput.decode(sys.getfilesystemencoding()).splitlines()
        if len(deletedDocs) > 0:
            self.callHg('remove', *deletedDocs)
        hgOutput = self.callHg('status', '--unknown', '--no-status', self.db.documents)
        newDocs = hgOutput.decode().splitlines()
        if len(newDocs) > 0:
            self.callHg('add', *newDocs)
        # TODO: sophisticated analysis of diff to current head
        self.callHg('commit', '--message', msg or 'Auto-Commit by BibTeX VCS')
        if self.hasRemote:
            self.callHg('push')
        self.db.reload()
        return True

    def revision(self):
        return self.callHg('log', '-l', '1', '--template', '{rev}\n{date|isodate}'
                           ).decode().splitlines()

    @classmethod
    def clone(cls, url, target, username=None, password=None):
        cls._callHg('clone', url, target, username=username, password=password)


class NoVCSInterface(VCSInterface):
    """Dummy VCS interface for the case of a database that is not under version control."""

    vcsType = None

    def __getattribute__(self, name):
        if name in ("add", "localChanges", "update", "commit"):
            return lambda *args, **kwargs: False
        return super().__getattribute__(name)

    def revision(self):
        return None, None



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
from bibtexvcs.bibfile import MacroReference, MONTHS
import sys, inspect

def performDatabaseCheck(database):
    me = sys.modules[__name__]
    checks = [fun for fname, fun in inspect.getmembers(me, inspect.isfunction) if fname.startswith('check')]
    errors = []
    for check in checks:
        errors.extend(list(check(database)))
    return errors

class CheckFailed(Exception):
    pass

def checkMacros(database):
    """Check if all macros referenced in the database exist in its journals file."""
    bib = database.bibfile
    for entry in bib.values():
        for field, value in entry.items():
            if isinstance(value, MacroReference):
                if value.name in bib.macroDefinitions or value.name in database.journals:
                    continue
                if value.name in MONTHS:
                    continue
                yield CheckFailed("The macro '{m}' used for field '{f}' in bibtex "
                                  "entry '{e}' is defined neither in the database nor "
                                  "in the journals file."
                                  .format(f=field, m=value.name, e=entry.citekey))


def checkFileLinks(database):
    """Check that the files linked to in the database match those existing in the `documents`
    directory. Additionally, check that all documents are contained in the `documents` directory.
    """

    dbFilesSet = set(database.referencedDocuments())
    fsFilesSet = set(database.existingDocuments())

    if len(dbFilesSet - fsFilesSet) > 0:
        yield CheckFailed("The following file(s) are referenced in the bibtex file but do not "
                          "exist in the documents directory:\n{}"
                          .format("\n".join(dbFilesSet - fsFilesSet)))
    if len(fsFilesSet - dbFilesSet) > 0:
        yield CheckFailed("The following file(s) in the documents directory are not referenced by "
                          "any entry in the bibtex file:\n{}"
                          .format("\n".join(fsFilesSet - dbFilesSet)))


def checkASCIIFilenames(database):
    """Check that all file names are ASCII. This is sensible because non-ASCII file names lead
    to problems with most VCS systems.
    """
    for filename in database.existingDocuments():
        try:
            filename.encode('ascii')
        except UnicodeEncodeError:
            yield CheckFailed('The file name "{}" contains non-ASCII characters.'.format(filename))


MONTHS = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]


def checkMonthMacros(database):
    """Checks that the ``month`` field only contains (proper) macros."""

    for entry in database.bibfile.values():
        if 'month' not in entry:
            continue
        month = entry['month']
        if isinstance(month, str):
            yield CheckFailed("Month field in entry '{e}' contains the string '{s}' instead "
                              "of a month macro.".format(e=entry.citekey, s=month))
        elif isinstance(month, MacroReference):
            if month.name not in MONTHS:
                yield CheckFailed("Invalid month macro '{}' used in entry '{}'"
                                 .format(month.name, entry.citekey))
        else:
            #  must be a list of macros
            if len(month) % 2 != 1:
                yield CheckFailed("Invalid month definition '{}' in '{}': Must be either a single "
                                 "month macro or of the format 'mar / apr'."
                                 .format(month, entry.citekey))
                continue
            separators = [ month[i] for i in range(1, len(month), 2) ]
            macros = [ month[i] for i in range(0, len(month), 2) ]
            for separator in separators:
                if separator.strip() != '/':
                    yield CheckFailed("Invalid month definition '{}' in '{}': Expected '/' but got "
                                      "'{}".format(month, entry.citekey, separator.strip()))
            for macro in macros:
                if not isinstance(macro, MacroReference) or macro.name not in MONTHS:
                    yield CheckFailed("Invalid month definition '{}' in '{}'"
                                      .format(month, entry.citekey))

def checkJabrefFileDirectory(database):
    identifier = 'jabref-meta: fileDirectory:'
    for comment in database.bibfile.comments:
        if comment.comment.startswith(identifier):
            if comment.comment[len(identifier):] != database.documents + ';':
                yield CheckFailed('JabRef fileDirectory field does not coincide with '
                                  'the configured one.')
